﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Threading;


// ForkEach attempts to make foreach parallel.
// As number of cores will reach double digits this type of construct
// will provide for a good error free multithreaded paradigm.
// Eventually it should become a part of a language or a compiler optimization for foreach().
// Compile with VS 2008
// Feel free to use ForkEach and modify it as long as this notice remains in the file.
// Original Version written by Andrey Makushkin in 2008.  Comments? toandrey(at)yahoo.com
// Please send me your improved modifications.
//    Use it like this:
//    string[] col = { "0", "1", "2", "3", "4", "5", "6", "7", "8", "9" };
//    ForkEach<string>.IN(col, delegate(string item) {
//       Console.Out.WriteLine(item); // see the output order
//    });


namespace ForkEach{

    // Affinity was taken out so that it has a single "static instance" :)
    // Since ForkEach is a generic, every type would have to run init() and keep own procCount
    // This would make init run for every new type. We only want it to run once
    static class Affinity {
        private static int procCount = 0;

        public static int ThreadCount {
            get{ return (procCount == 0)? init() : procCount; }
            set{ if(value > 0){ procCount = value; } } // go ahead, shoot yourself in a foot
        }

        // this method tries to figure out the best number of threads forkeach should use
        // we're trying to prevent 1000 User Work Items queued in ThreadPool
        private static int init() {
            long mask = (long)System.Diagnostics.Process.GetCurrentProcess().ProcessorAffinity;
            procCount = 0; // extract the number of processors from the mask
            for (int i = 0; i < 64; ++i) {
                long lsb = (mask >> i) & 1;
                procCount += (int)lsb; // count how many bits are 1
            }
            procCount = procCount * 2; // double the # of ProcessorAffinity (50% of the threads will sleep)
            int workers, completion;
            ThreadPool.GetMaxThreads(out workers, out completion);
            procCount = (procCount > workers) ? workers : procCount; // should not be true, but just in case
            return procCount;
        }
    } // class Affinity


    class ForkEach<T>{
        private static int procCount = Affinity.ThreadCount; // yes, you could take it out and put a magic #
        private bool isLocked = false;
        private List<Exception> savedExceptions = null;
        private Semaphore sem = null;
        private GenDelegate delt = null;
        public delegate void GenDelegate(T t);

        private ForkEach(GenDelegate initDel, Semaphore sharedSem, bool locked, List<Exception> exceptions){
            delt = initDel;
            sem = sharedSem;
            isLocked = locked;
            savedExceptions = exceptions;
        }

        private void run(object obj) {
            try{
                if(isLocked){
                    lock(obj){
                        delt( (T)obj );
                    }
                } else {
                    delt( (T)obj );
                }
            } catch (Exception ex) { // save the exception, swallow it here and rethrow in the main thread
                lock(savedExceptions){
                    savedExceptions.Add(ex);
                }
            } finally {
                sem.Release();
            }
        }

        public static void IN(IEnumerable<T> col,GenDelegate del){ IN(col,false,del); } // __forceinline JIT :)
        public static void IN(IEnumerable<T> col, bool locked, GenDelegate del) {
            if(del == null){throw new ArgumentException("Delegate can not be null","del"); }
            List<Exception> exceptions = new List<Exception>();
            Semaphore semaphore = new Semaphore(procCount, procCount);  // do not use named semaphore for all to share!

            foreach (T item in col) {
                if(exceptions.Count > 0){ break; } // one of the threads generated an exception
                semaphore.WaitOne();    // control how many threads we are running
                ForkEach<T> fork = new ForkEach<T>(del, semaphore, locked, exceptions);
                ThreadPool.QueueUserWorkItem(fork.run, item); // run delegate on the next available thread
            }

            for(int i = 0; i < procCount; i++){
                semaphore.WaitOne(); // wait for all threads to finish
            }
            semaphore.Close();

            if (exceptions.Count > 0) { // one of the threads generated an exception
                string msg = String.Format("{0} exceptions caught in ForkEach.IN(). See inner exception for more details.",exceptions.Count);
                Exception ex = new Exception(msg,exceptions[0]);
                throw ex;
            }
        }

    } // class ForkEach
} // namespace
